"""
Utilities for image manipulation.
"""
from __future__ import absolute_import

import base64
import io
import math
from typing import Text, Union

from PIL import Image

from applitools.common import logger
from applitools.common.errors import EyesError
from applitools.common.geometry import Point, Region
from applitools.common.scale_provider import ScaleProvider

from . import argument_guard
from .converters import round_converter

__all__ = (
    "image_from_path",
    "image_from_bytes",
    "image_from_base64",
    "scale_image",
    "get_base64",
    "get_bytes",
    "get_image_part",
)


def image_from_path(path_or_image):
    # type: (Union[Image.Image, Text]) -> Image.Image
    """
    Reads the PNG data from the given file stream and returns a new Image instance.
    """
    if isinstance(path_or_image, Image.Image):
        return path_or_image
    return Image.open(path_or_image)


def image_from_bytes(png_bytes):
    # type: (bytes) -> Image.Image
    """
    Reads the PNG data from the given png bytes and returns a new Image instance.

    :param png_bytes: Png bytes.
    :return: Image instance.
    """
    return Image.open(io.BytesIO(png_bytes))


def image_from_base64(base64_str):
    # type: (str) -> Image.Image
    """
    Reads the PNG data from the given png bytes and returns a new Image instance.

    :param png_bytes: Png bytes.
    :return: Image instance.
    """
    return Image.open(io.BytesIO(base64.b64decode(base64_str)))


def scale_image(image, scale):
    # type: (Image.Image, Union[float, ScaleProvider]) -> Image.Image
    argument_guard.is_a(image, Image.Image)
    argument_guard.not_none(scale_image)

    if isinstance(scale, float) or isinstance(scale, int):
        scale_ratio = scale
    else:
        scale_ratio = scale.scale_ratio

    if scale_ratio == 1:
        return image

    image_ratio = float(image.height) / float(image.width)
    scale_width = int(math.ceil(image.width * scale_ratio))
    scale_height = int(math.ceil(scale_width * image_ratio))
    image = image.convert("RGBA")
    scaled_image = image.resize((scale_width, scale_height), resample=Image.BICUBIC)
    return scaled_image


def get_base64(image):
    # type: (Image.Image) -> str
    """
    Gets the base64 representation of the PNG bytes.

    :return: The base64 representation of the PNG bytes.
    """
    image_bytes_stream = io.BytesIO()
    image.save(image_bytes_stream, format="PNG")
    image64 = base64.b64encode(image_bytes_stream.getvalue()).decode("utf-8")
    image_bytes_stream.close()
    return image64


def get_bytes(image):
    # type: (Image.Image) -> bytes
    """
    Gets the image bytes.

    :return: The image bytes.
    """
    image_bytes_stream = io.BytesIO()
    image.save(image_bytes_stream, format="PNG")
    image_bytes = image_bytes_stream.getvalue()
    image_bytes_stream.close()
    return image_bytes


def get_image_part(image, region):
    # type: (Image.Image, Region) -> Image.Image
    """
    Get a copy of the part of the image given by region.

    :return: The part of the image.
    """
    if region.is_empty:
        raise EyesError("region is empty!")
    return image.crop(box=(region.left, region.top, region.right, region.bottom))


def save_image(image, filename):
    try:
        logger.info("Saving file: {}".format(filename))
        image.save(filename)
    except Exception:
        logger.warning("Failed to save image")


def crop_image(image, region_to_crop):
    # type: (Image.Image, Region) -> Image.Image
    argument_guard.is_a(image, Image.Image)
    argument_guard.is_a(region_to_crop, Region)

    image_region = Region.from_(image)
    image_region = image_region.intersect(region_to_crop)
    if image_region.is_size_empty:
        logger.warning(
            "requested cropped area results in zero-size image! "
            "Cropped not performed. Returning original image."
        )
        return image

    if image_region != region_to_crop:
        logger.warning("requested cropped area overflows image boundaries.")

    cropped_image = image.crop(
        box=(
            image_region.left,
            image_region.top,
            image_region.right,
            image_region.bottom,
        )
    )
    return cropped_image


def paste_image(base_image, part_image, position):
    # type: (Image.Image, Image.Image, Point) -> Image.Image
    argument_guard.is_a(base_image, Image.Image)
    argument_guard.is_a(part_image, Image.Image)
    argument_guard.is_a(position, Point)

    try:
        base_image.paste(part_image, box=(position.x, position.y))
    except TypeError as e:
        logger.exception(e)
        logger.debug("Convert position values to integer")
        base_image.paste(
            part_image, box=(round_converter(position.x), round_converter(position.y))
        )
    return base_image


def rotate_image(image, rotate):
    # type: (Image.Image, int) -> Image.Image
    return image.rotate(rotate)


def crop_to_specific_size_if_needed(image, max_image_height, max_image_area):
    # type: (Image, int, int) -> Image
    if (
        image.height <= max_image_height
        and image.width * image.height <= max_image_area
    ):
        return image

    trimmed_height = min(max_image_area / image.width, max_image_height)
    new_region = Region(0, 0, image.width, trimmed_height)
    if new_region.is_size_empty:
        return image

    image = crop_image(image, new_region)
    return image



import attr


class DictAccessMixin(object):
    """Make dict-like object from attrs class"""

    def __getitem__(self, item):
        fields = [a.name for a in attr.fields(self.__class__)]
        if isinstance(item, int):
            item = fields[item]
        if item not in fields:
            raise KeyError("item: {}, fields: {}".format(item, fields))
        return getattr(self, item)



"""
Logs handling.
"""
from __future__ import absolute_import

import cgitb
import inspect
import logging
import logging.config
import platform
import re
import sys
import threading
import typing as tp
from enum import Enum
from typing import Any, Dict, Optional, Text

import attr
import structlog
import structlog.dev

from applitools.common.utils import json_utils
from applitools.common.utils.general_utils import get_env_with_prefix

__all__ = ("StdoutLogger", "FileLogger")
_DEFAULT_HANDLER_LEVEL = int(get_env_with_prefix("LOGGER_LEVEL", logging.INFO))
_frames_regex = re.compile(
    r"function calls leading up to the error, in the order they occurred\.\s*"
    r"(.*?)"
    r"\s*The above is a description of an error in a Python program.",
    re.DOTALL,
)


class Stage(Enum):
    GENERAL = "GENERAL"
    OPEN = "OPEN"
    CHECK = "CHECK"
    CLOSE = "CLOSE"
    RENDER = "RENDER"
    RESOURCE_COLLECTION = "RESOURCE_COLLECTION"
    LOCATE = "LOCATE"
    OCR = "OCR"


def create_message_from_log(
    agent_id,  # type: Text
    test_id,  # type: Text
    stage,  # type: Stage
    data,  # type: Dict[Text, Any]
    type=None,  # type: Optional[Text]
    methods_back=3,  # type: int
):
    # type: (...) -> Text
    d = {
        "agentId": agent_id,
        "testId": test_id,
        "stage": stage,
        "threadId": threading.current_thread().name,
        "stackTrace": [inspect.stack()[i][3] for i in range(2, methods_back + 1)],
        "pythonVersion": "{} {}".format(
            platform.python_implementation(),
            platform.python_version(),
        ),
        "platformName": platform.platform(),
    }
    if type:
        d["type"] = type

    d.update(data)
    return json_utils.to_json(d)


class StdoutLogger(object):
    """
    A simple logger class for printing to STDOUT.
    """

    @tp.overload
    def __init__(self, is_verbose):
        # type: (bool) -> None
        """
        :param is_verbose: enables logging DEBUG messages
        """
        pass

    @tp.overload
    def __init__(self, name, level):
        # type: (tp.Text, int) -> None
        """
        :param name: unused
        :param level: log level (e.g. logging.WARNING)
        """
        pass

    def __init__(self, name=None, level=_DEFAULT_HANDLER_LEVEL, is_verbose=None):
        is_verbose = is_verbose if is_verbose is not None else name
        if is_verbose is True:
            self.level = logging.DEBUG
        elif is_verbose is False:
            self.level = logging.INFO
        else:
            self.level = level

    def configure(self, std_logger):
        # type: (logging.Logger) -> None
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(self.level)
        handler.setFormatter(
            structlog.stdlib.ProcessorFormatter(
                structlog.dev.ConsoleRenderer(), _pre_chain
            )
        )
        std_logger.addHandler(handler)


@attr.s
class FileLogger(object):
    """
    A simple logger class for outputting log messages to a file

    :param filename: The name of this file to which logs should be written.
    :param mode: The mode in which the log file is opened
                 ('a' for appending, 'w' for overwrite).
    :param encoding: The encoding in which logs will be written to the file.
    :param delay: If True, file will not be opened until the first log message
                  is emitted.
    :param name: unused
    :param level: The log level (e.g., logging.DEBUG)
    """

    filename = attr.ib(default="eyes.log")
    mode = attr.ib(default="a")
    encoding = attr.ib(default=None)
    delay = attr.ib(default=0)
    name = attr.ib(default=None)
    level = attr.ib(default=_DEFAULT_HANDLER_LEVEL)  # type: int

    def configure(self, std_logger):
        # type: (logging.Logger) -> None
        handler = logging.FileHandler(
            self.filename, self.mode, self.encoding, self.delay
        )
        handler.setLevel(self.level)
        handler.setFormatter(
            structlog.stdlib.ProcessorFormatter(
                structlog.processors.JSONRenderer(), _pre_chain
            )
        )
        std_logger.addHandler(handler)


def set_logger(logger=None):
    # type: (tp.Union[StdoutLogger, FileLogger]) -> None
    std_logger = logging.getLogger(__name__)
    logger.configure(std_logger)


def _add_thread_name(_, __, event_dict):
    event_dict["_thread_name"] = threading.current_thread().name
    return event_dict


def _format_exc_stack_trace_with_vars(_, __, event_dict):
    exc_info = event_dict.get("exc_info", None)
    if exc_info:
        if sys.version_info[0] >= 3 and isinstance(exc_info, BaseException):
            exc_info = (exc_info.__class__, exc_info, exc_info.__traceback__)
        elif isinstance(exc_info, tuple):
            pass
        elif exc_info:
            exc_info = sys.exc_info()
        try:
            match = _frames_regex.search(cgitb.text(exc_info))
            event_dict["stack_trace_with_vars"] = match[1]
        except Exception as e:
            event_dict["stack_trace_with_vars"] = "Collection failure: {}".format(e)
    return event_dict


_timestamper = structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S.%f", utc=False)
_pre_chain = [
    structlog.stdlib.add_log_level,
    _timestamper,
    _add_thread_name,
]
structlog.configure(
    processors=_pre_chain
    + [
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.StackInfoRenderer(),
        _format_exc_stack_trace_with_vars,
        structlog.processors.format_exc_info,
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ],
    context_class=dict,
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    cache_logger_on_first_use=True,
)

# Allow everything so handlers can filter on their levels
logging.getLogger(__name__).setLevel(logging.DEBUG)
logging.getLogger(__name__).propagate = False

_logger = structlog.get_logger().bind()
bind = _logger.bind
info = _logger.info
debug = _logger.debug
warning = _logger.warning
exception = _logger.exception
error = _logger.error



from __future__ import absolute_import, unicode_literals

import json
import re
from typing import Dict, Text


def efficient_string_replace(ref_id_open_token, ref_id_end_token, input, replacements):
    # type: (Text, Text, Text, Dict[Text, Text]) -> Text
    r = re.compile(re.escape(ref_id_open_token) + "(.*?)" + re.escape(ref_id_end_token))

    def replacement(match):
        replacement.performed = True  # no nonlocal keyword in py2
        return replacements[match.group(1)]

    for _ in range(len(replacements) + 1):  # limit retries to avoid endless loop
        replacement.performed = False
        input = r.sub(replacement, input)
        if not replacement.performed:
            break
    else:
        raise RuntimeError("Cyclic replacement pattern found")

    return input


def clean_for_json(s):
    # type: (Text) -> Text
    # make json array and remove [" and "]
    return json.dumps([s])[2:-2]



from __future__ import absolute_import

from typing import TYPE_CHECKING, Any, Dict, List, Text, Tuple, Union

if TYPE_CHECKING:
    from appium.webdriver import WebElement as AppiumWebElement
    from selenium.webdriver.remote.webdriver import WebDriver
    from selenium.webdriver.remote.webelement import WebElement as SeleniumWebElement

    from applitools.common.geometry import RectangleSize, Region
    from applitools.core.triggers import ActionTrigger
    from applitools.selenium.webdriver import EyesWebDriver
    from applitools.selenium.webelement import EyesWebElement

    ViewPort = Union[Dict[Text, int], RectangleSize]  # typedef
    CodedRegionPadding = Dict[Text, int]  # typedef
    Num = Union[int, float]

    AnyWebDriver = Union[EyesWebDriver, WebDriver]  # typedef
    AnyWebElement = Union[
        EyesWebElement, SeleniumWebElement, AppiumWebElement
    ]  # typedef

    FrameNameOrId = Text  # typedef
    FrameIndex = int  # typedef
    FrameReference = Union[FrameNameOrId, FrameIndex, AnyWebElement]  # typedef

    # could contain MouseTrigger, TextTrigger
    UserInputs = List[ActionTrigger]  # typedef
    RegionOrElement = Union[EyesWebElement, Region]  # typedef

    SeleniumBy = Text  # typedef
    BySelector = List[SeleniumBy, Text]  # typedef
    CssSelector = Text  # typedef
    REGION_VALUES = Union[Region, CssSelector, AnyWebElement, BySelector]  # typedef
    FLOATING_VALUES = Union[Region, CssSelector, AnyWebElement, BySelector]  # typedef



import typing
from threading import Lock
from typing import Dict, Optional

from applitools.common.geometry import RectangleSize
from applitools.common.ultrafastgrid.config import (
    DeviceName,
    IosDeviceName,
    ScreenOrientation,
)

if typing.TYPE_CHECKING:
    from applitools.core import ServerConnector


_lock = Lock()
_devices_sizes = {}  # type: Dict[type, Dict[ScreenOrientation, RectangleSize]]
_server_connector = None  # type: Optional[ServerConnector]


def maybe_set_server_connector(server_connector):
    with _lock:
        global _server_connector
        if not _devices_sizes and not _server_connector:
            _server_connector = server_connector


def size(device_name, screen_orientation):
    _maybe_fetch_device_sizes()
    return _devices_sizes[device_name][screen_orientation]


def _maybe_fetch_device_sizes():
    with _lock:
        if not _devices_sizes:
            global _server_connector
            for dev_type in DeviceName, IosDeviceName:
                _devices_sizes.update(_fetch_devices_sizes(dev_type))
            _server_connector = None


def _fetch_devices_sizes(device_type):
    sizes = _server_connector.get_devices_sizes(device_type)
    return {
        device: {
            orientation: RectangleSize.from_(sizes[device.value][orientation.value])
            for orientation in ScreenOrientation
        }
        for device in device_type
    }



import typing

import attr

from .app_output import AppOutput
from .match import ImageMatchSettings
from .utils.json_utils import JsonInclude

if typing.TYPE_CHECKING:
    from typing import Optional, Text

    from .utils.custom_types import UserInputs


__all__ = ("MatchWindowData", "Options")


@attr.s
class Options(object):
    """
    Encapsulates the "Options" section of the MatchExpectedOutput body data.

    :ivar name: The tag of the window to be matched.
    :ivar user_nputs: A list of triggers between the previous matchWindow call and
    the current matchWindow call. Can be array of size 0, but MUST NOT be null.
    :ivar ignore_mismatch: Tells the server whether or not to store a mismatch for
    the current window as window in the session.
    :ivar ignore_match: Tells the server whether or not to store a match for the
    current window as window in the session.
    :ivar force_mismatch: Forces the server to skip the comparison process and mark
    the current window as a mismatch.
    :ivar force_match: Forces the server to skip the comparison process and mark the
    current window as a match.
    :ivar image_match_settings
    :ivar source: App name or web page domain where checking is performed
    """

    name = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    user_inputs = attr.ib(metadata={JsonInclude.THIS: True})  # type: UserInputs
    replace_last = attr.ib(metadata={JsonInclude.THIS: True})  # type: bool
    ignore_mismatch = attr.ib(metadata={JsonInclude.THIS: True})  # type: bool
    ignore_match = attr.ib(metadata={JsonInclude.THIS: True})  # type: bool
    force_mismatch = attr.ib(metadata={JsonInclude.THIS: True})  # type: bool
    force_match = attr.ib(metadata={JsonInclude.THIS: True})  # type: bool
    image_match_settings = attr.ib(
        type=ImageMatchSettings, metadata={JsonInclude.THIS: True}
    )  # type: ImageMatchSettings
    source = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Optional[Text]
    render_id = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Optional[Text]
    variant_id = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Optional[Text]


@attr.s
class MatchWindowData(object):
    """
    Encapsulates the data to be sent to the agent on a "match_window" command.

    :param user_inputs: A list of triggers between the previous
    matchWindow call and the current matchWindow call. Can be array of size 0,
    but MUST NOT be null.
    :param app_output: The appOutput for the current matchWindow call.
    :param tag: The tag of the window to be matched.
    :param ignore_mismatch: A flag indicating whether the server should ignore the image
     in case of a mismatch.
    :param options: A set of match options for the server.
    :param agent_setup: An object representing the configuration used to create the image.
    """

    # TODO Remove redundancy: user_inputs and ignore_mismatch should only be inside
    #  Options. (requires server version update).
    ignore_mismatch = attr.ib(metadata={JsonInclude.THIS: True})  # type: bool
    user_inputs = attr.ib(metadata={JsonInclude.THIS: True})  # type: UserInputs
    app_output = attr.ib(
        type=AppOutput, metadata={JsonInclude.THIS: True}
    )  # type: AppOutput
    tag = attr.ib(metadata={JsonInclude.THIS: True})  # type: Text
    options = attr.ib(type=Options, metadata={JsonInclude.THIS: True})  # type: Options
    agent_setup = attr.ib(metadata={JsonInclude.THIS: True})  # type: Optional[Text]
    render_id = attr.ib(metadata={JsonInclude.NON_NONE: True})  # type: Optional[Text]



from __future__ import absolute_import

from datetime import datetime
from enum import Enum
from typing import Text, Union

from .compat import quote_plus


def value_from_enum(e):
    #  type: (Union[Enum, str]) -> str
    return e.value if isinstance(e, Enum) else e


def isoformat(d):
    # type: (datetime) -> str
    return d.isoformat()


def round_converter(x):
    # type: (float) -> int
    return int(round(x))


def str2bool(v):
    # type: (Text) -> bool
    if v is None:
        return False
    return v.lower() in ("yes", "true", "t", "1")
